#!/usr/bin/python
import math
import os.path as op
import sys
import time  # noqa
from datetime import datetime

import pyqtgraph as pg
from PyQt5 import QtCore, QtGui, QtWidgets

from kite.qt_utils import QRangeSlider, SceneLog, loadUi, validateFilename
from kite.scene import Scene
from kite.scene_stack import SceneStack

from .base import get_resource
from .scene_model import SceneModel
from .tab_aps import KiteAPS  # noqa
from .tab_covariance import KiteCovariance  # noqa
from .tab_quadtree import KiteQuadtree  # noqa
from .tab_scene import KiteScene


class Spool(QtWidgets.QApplication):
    def __init__(self, scene=None, import_file=None, load_file=None):
        super().__init__(["Spool"])
        # self.setStyle('plastique')
        splash_img = QtGui.QPixmap(get_resource("spool_splash.png")).scaled(
            QtCore.QSize(400, 250), QtCore.Qt.KeepAspectRatio
        )
        self.splash = QtWidgets.QSplashScreen(
            splash_img, QtCore.Qt.WindowStaysOnTopHint
        )
        self.updateSplashMessage("Scene")
        self.splash.show()
        self.processEvents()

        self.spool_win = SpoolMainWindow()
        self.spool_win.sigLoadingModule.connect(self.updateSplashMessage)

        self.spool_win.actionExit.triggered.connect(self.exit)
        self.aboutToQuit.connect(self.spool_win.model.worker_thread.quit)

        if scene is not None:
            self.addScene(scene)
        elif import_file is not None:
            self.importScene(import_file)
        elif load_file is not None:
            self.loadScene(load_file)

        self.spool_win.show()
        self.splash.finish(self.spool_win)

    @QtCore.pyqtSlot(str)
    def updateSplashMessage(self, msg=""):
        self.splash.showMessage(f"Loading {msg.title()} ...", QtCore.Qt.AlignBottom)

    def addScene(self, scene):
        self.spool_win.addScene(scene)

    def importScene(self, filename):
        self.spool_win.model.importFile(filename)

    def loadScene(self, filename):
        self.spool_win.model.loadFile(filename)


class SpoolMainWindow(QtWidgets.QMainWindow):
    VIEWS = [KiteScene, KiteQuadtree, KiteCovariance, KiteAPS]

    sigImportFile = QtCore.pyqtSignal(str)
    sigLoadFile = QtCore.pyqtSignal(str)
    sigLoadConfig = QtCore.pyqtSignal(str)
    sigExportWeightMatrix = QtCore.pyqtSignal(str)
    sigLoadingModule = QtCore.pyqtSignal(str)

    def __init__(self, *args, **kwargs):
        QtWidgets.QMainWindow.__init__(self, *args, **kwargs)
        loadUi(get_resource("spool.ui"), baseinstance=self)

        self.views = []
        self.active_view = None

        self.ptree = KiteParameterTree(showHeader=False)
        self.ptree.setMinimumWidth(400)
        self.dock_ptree = QtWidgets.QDockWidget("Parameters", self)
        self.dock_ptree.setFeatures(
            QtWidgets.QDockWidget.DockWidgetMovable
            | QtWidgets.QDockWidget.DockWidgetFloatable
        )
        self.dock_ptree.setWidget(self.ptree)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.dock_ptree)

        self.tabs = QtWidgets.QTabWidget(self)
        self.dock_tabs = QtWidgets.QDockWidget(self)
        self.dock_tabs.setTitleBarWidget(QtWidgets.QWidget())
        self.dock_tabs.setWidget(self.tabs)
        self.dock_tabs.setFeatures(QtWidgets.QDockWidget.NoDockWidgetFeatures)

        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.dock_tabs)
        self.setCentralWidget(self.dock_tabs)

        self.model = SceneModel(self)
        self.model.sigSceneModelChanged.connect(self.buildViews)

        # Connecting signals
        self.sigLoadFile.connect(self.model.loadFile)
        self.sigImportFile.connect(self.model.importFile)
        self.sigLoadConfig.connect(self.model.loadConfig)
        self.sigExportWeightMatrix.connect(self.model.exportWeightMatrix)

        self.actionSave_config.triggered.connect(self.onSaveConfig)
        self.actionSave_scene.triggered.connect(self.onSaveScene)
        self.actionLoad_config.triggered.connect(self.onLoadConfig)
        self.actionLoad_scene.triggered.connect(self.onOpenScene)

        self.actionImport_scene.triggered.connect(self.onImportScene)
        self.actionExport_quadtree.triggered.connect(self.onExportQuadtree)
        self.actionExport_weights.triggered.connect(self.onExportWeightMatrix)

        self.actionAbout_Spool.triggered.connect(self.aboutDialog().show)
        self.actionAbout_Qt5.triggered.connect(
            lambda: QtWidgets.QMessageBox.aboutQt(self)
        )
        self.actionHelp.triggered.connect(
            lambda: QtGui.QDesktopServices.openUrl(QtCore.QUrl("https://pyrocko.org"))
        )

        self.log = SceneLog(self, self.model)
        self.actionLog.triggered.connect(self.log.show)

        self.progress = QtWidgets.QProgressDialog("", None, 0, 0, self)
        self.progress.setValue(0)
        self.progress.closeEvent = lambda ev: ev.ignore()
        self.progress.setMinimumWidth(400)
        self.progress.reset()
        self.progress_timer = None

        self.state_hash = None

    def aboutDialog(self):
        self._about = QtWidgets.QDialog(self)
        loadUi(get_resource("about.ui"), baseinstance=self._about)
        return self._about

    def addScene(self, scene):
        self.model.setScene(scene)
        self.buildViews()
        self.state_hash = scene.get_state_hash()

    def buildViews(self):
        scene = self.model.getScene()

        title = scene.meta.filename or "Untitled"
        self.setWindowTitle(f"Spool - {title}")
        if scene is None or self.tabs.count() != 0:
            return
        for v in self.VIEWS:
            self.addView(v)
        self.model.sigProgressStarted.connect(
            self.progressStarted, type=QtCore.Qt.QueuedConnection
        )
        self.model.sigProgressFinished.connect(
            self.progressFinished, type=QtCore.Qt.QueuedConnection
        )

        self.tabs.currentChanged.connect(self.activateView)

        if isinstance(scene, SceneStack):
            self.addTimeSlider()

        self.activateView(0)

    def addView(self, view):
        self.sigLoadingModule.emit(view.title)
        view = view(self)
        self.tabs.addTab(view, view.title)

        if hasattr(view, "parameters"):
            for parameter in view.parameters:
                self.ptree.addParameters(parameter)

        self.views.append(view)

    @QtCore.pyqtSlot(int)
    def activateView(self, index):
        if self.active_view is not None:
            self.active_view.deactivateView()

        self.active_view = self.views[index]
        self.active_view.activateView()

    @QtCore.pyqtSlot(object)
    def progressStarted(self, args):
        if self.progress.isVisible():
            return

        nargs = len(args)

        text = args[0]
        maximum = 0 if nargs < 2 else args[1]
        progress_func = None if nargs < 3 else args[2]

        self.progress.setWindowTitle("Processing...")
        self.progress.setLabelText(text)
        self.progress.setMaximum(maximum)
        self.progress.setValue(0)

        @QtCore.pyqtSlot()
        def updateProgress():
            if progress_func:
                self.progress.setValue(progress_func())

        self.progress_timer = QtCore.QTimer()
        self.progress_timer.timeout.connect(updateProgress)
        self.progress_timer.start(250)

        self.progress.show()

    @QtCore.pyqtSlot()
    def progressFinished(self):
        self.progress.reset()
        if self.progress_timer is not None:
            self.progress_timer.stop()
            self.progress_timer = None

    def getSceneDirname(self):
        if self.model.scene.meta.filename:
            return op.dirname(self.model.scene.meta.filename)
        return None

    def onSaveConfig(self):
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            parent=self,
            directory=self.getSceneDirname(),
            filter="YAML file *.yml (*.yml)",
            caption="Save scene YAML config",
        )
        if not validateFilename(filename):
            return
        self.model.scene.saveConfig(filename)
        QtWidgets.QMessageBox.information(
            self,
            "Scene config saved",
            "Scene config successfully saved!"
            f'<p style="font-family: monospace;">{filename}'
            "</p>",
        )
        self.state_hash = self.model.scene.get_state_hash()

    def onSaveScene(self):
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            parent=self,
            directory=self.getSceneDirname(),
            filter="YAML *.yml and NumPy container *.npz (*.yml *.npz)",
            caption="Save scene",
        )
        if not validateFilename(filename):
            return
        self.model.scene.save(filename)
        QtWidgets.QMessageBox.information(
            self,
            "Scene saved",
            "Scene successfully saved!"
            '<p style="font-family: monospace;">%s'
            "</p>" % filename,
        )
        self.state_hash = self.model.scene.get_state_hash()

    def onLoadConfig(self):
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            filter="YAML file *.yml (*.yml)", caption="Load scene YAML config"
        )
        if not validateFilename(filename):
            return
        self.sigLoadConfig.emit(filename)

    def onOpenScene(self):
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            parent=self,
            filter="YAML *.yml and NumPy container *.npz (*.yml *.npz)",
            caption="Load kite scene",
        )
        if not validateFilename(filename):
            return
        self.model.loadFile(filename)

    def onImportScene(self):
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            filter="Supported Formats, *.grd, *.geo, *unw*, *.mat "
            "(*.grd,*.geo,*unw*,*.mat);;"
            "GMT5SAR Scene, *.grd (*.grd);;"
            "ISCE Scene, *unw* (*unw*);;Gamma Scene *.geo (*.geo);;"
            "Matlab Container, *.mat (*.mat);;Any File * (*)",
            caption="Import scene to spool",
        )
        if not validateFilename(filename):
            return
        self.sigImportFile.emit(filename)

    def onExportQuadtree(self):
        filename, flt = QtWidgets.QFileDialog.getSaveFileName(
            parent=self,
            filter="GeoJSON (*.geojson *.json);;CSV File *.csv (*.csv)",
            caption="Export Quadtree",
        )

        if not validateFilename(filename):
            return

        if flt == "GeoJSON (*.geojson *.json)":
            if not filename.endswith(".geojson") and not filename.endswith(".json"):
                filename += ".geojson"
            self.model.quadtree.export_geojson(filename)

        elif flt == "CSV File *.csv (*.csv)":
            if not filename.endswith(".csv"):
                filename += ".geojson"
            self.model.quadtree.export_csv(filename)
        else:
            raise ValueError("unknown filter")

    def onExportWeightMatrix(self):
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            filter="Text File *.txt (*.txt)",
            caption="Export Covariance Weights",
        )
        if not validateFilename(filename):
            return
        self.sigExportWeightMatrix.emit(filename)

    def addTimeSlider(self):
        stack = self.model.getScene()

        self.time_slider = QRangeSlider(self)
        self.time_slider.setMaximumHeight(50)

        slider_tmin = math.ceil(stack.tmin)
        slider_tmax = math.floor(stack.tmax)

        def datetime_formatter(value):
            return datetime.fromtimestamp(value).strftime("%Y-%m-%d")

        self.time_slider.setMin(slider_tmin)
        self.time_slider.setMax(slider_tmax)
        self.time_slider.setRange(slider_tmin, slider_tmax)
        self.time_slider.setFormatter(datetime_formatter)

        @QtCore.pyqtSlot(int)
        def changeTimeRange():
            tmin, tmax = self.time_slider.getRange()
            stack.set_time_range(tmin, tmax)

        self.time_slider.startValueChanged.connect(changeTimeRange)
        self.time_slider.endValueChanged.connect(changeTimeRange)

        self.dock_time_slider = QtWidgets.QDockWidget(
            "Displacement time series - range control", self
        )
        self.dock_time_slider.setWidget(self.time_slider)
        self.dock_time_slider.setFeatures(QtWidgets.QDockWidget.DockWidgetMovable)
        self.dock_time_slider.setAllowedAreas(
            QtCore.Qt.BottomDockWidgetArea | QtCore.Qt.TopDockWidgetArea
        )

        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.dock_time_slider)

    def closeEvent(self, ev):
        if self.state_hash == self.model.scene.get_state_hash():
            return

        msg_box = QtWidgets.QMessageBox(parent=self, text="The scene has been modified")
        msg_box.setStandardButtons(
            QtWidgets.QMessageBox.Save
            | QtWidgets.QMessageBox.Discard
            | QtWidgets.QMessageBox.Cancel
        )
        msg_box.setInformativeText("Do you want to save your changes?")
        msg_box.setDefaultButton(QtWidgets.QMessageBox.Save)
        ret = msg_box.exec()

        if ret == QtWidgets.QMessageBox.Save:
            self.onSaveScene()
        elif ret == QtWidgets.QMessageBox.Cancel:
            ev.ignore()


class KiteParameterTree(pg.parametertree.ParameterTree):
    pass


def spool(*args, **kwargs):
    spool_app = Spool(*args, **kwargs)
    spool_app.exec_()
    spool_app.quit()


__all__ = ["Spool", "spool"]

if __name__ == "__main__":
    from kite.scene import SceneSynTest

    if len(sys.argv) > 1:
        sc = Scene.load(sys.argv[1])
    else:
        sc = SceneSynTest.createGauss()

    Spool(scene=sc)



import logging
from datetime import datetime

from PyQt5 import QtCore
from pyqtgraph import SignalProxy

try:
    from pyrocko.dataset.util import set_download_callback as pyrocko_download_callback
except ImportError:
    pyrocko_download_callback = None

from kite import Scene
from kite.qt_utils import SceneLogModel


class SceneModel(QtCore.QObject):
    """Proxy for :class:`kite.Scene` so we can change the scene"""

    sigSceneModelChanged = QtCore.pyqtSignal(object)

    sigSceneChanged = QtCore.pyqtSignal()
    sigConfigChanged = QtCore.pyqtSignal()

    sigFrameChanged = QtCore.pyqtSignal()
    sigQuadtreeChanged = QtCore.pyqtSignal()
    _sigQuadtreeChanged = QtCore.pyqtSignal()
    sigQuadtreeConfigChanged = QtCore.pyqtSignal()
    sigCovarianceChanged = QtCore.pyqtSignal()
    sigCovarianceConfigChanged = QtCore.pyqtSignal()

    sigProgressStarted = QtCore.pyqtSignal(object)
    sigProgressFinished = QtCore.pyqtSignal()
    sigCalculateWeightMatrixFinished = QtCore.pyqtSignal(object)

    sigHasElevation = QtCore.pyqtSignal()
    sigLogRecord = QtCore.pyqtSignal(object)

    def __init__(self, spool):
        QtCore.QObject.__init__(self)

        self.spool = spool
        self.scene = None
        self.frame = None
        self.quadtree = None
        self.covariance = None
        self.aps = None

        self.log = SceneLogModel(self)

        self._ = SignalProxy(
            self._sigQuadtreeChanged,
            rateLimit=10,
            delay=0,
            slot=lambda: self.sigQuadtreeChanged.emit(),
        )

        self._log_handler = logging.Handler()
        self._log_handler.setLevel(logging.DEBUG)
        self._log_handler.emit = self.sigLogRecord.emit

        logging.root.addHandler(self._log_handler)

        self._download_status = None
        if pyrocko_download_callback:
            pyrocko_download_callback(self.download_progress)

        self.qtproxy = QSceneQuadtreeProxy(self)

        self.worker_thread = QtCore.QThread()
        self.moveToThread(self.worker_thread)
        self.worker_thread.start()

    def setScene(self, scene):
        self.disconnectSlots()

        self.scene = scene
        self.frame = scene.frame
        self.quadtree = scene.quadtree
        self.covariance = scene.covariance
        self.aps = scene.aps

        self.connectSlots()
        self.sigSceneModelChanged.emit(object)

    def getScene(self):
        return self.scene

    def disconnectSlots(self):
        if self.scene is None:
            return

        self.scene.evChanged.unsubscribe(self.sigSceneChanged.emit)
        self.scene.evConfigChanged.unsubscribe(self.sigConfigChanged.emit)

        self.scene.frame.evChanged.unsubscribe(self.sigFrameChanged.emit)

        self.quadtree.evChanged.unsubscribe(self._sigQuadtreeChanged.emit)
        self.quadtree.evConfigChanged.unsubscribe(self.sigQuadtreeConfigChanged.emit)

        self.covariance.evChanged.unsubscribe(self.sigCovarianceChanged.emit)
        self.covariance.evConfigChanged.unsubscribe(
            self.sigCovarianceConfigChanged.emit
        )

        self.aps.evChanged.unsubscribe(self.sigAPSChanged.emit)

    def connectSlots(self):
        self.scene.evChanged.subscribe(self.sigSceneChanged.emit)
        self.scene.evConfigChanged.subscribe(self.sigCovarianceConfigChanged.emit)

        self.scene.frame.evChanged.subscribe(self.sigFrameChanged.emit)

        self.quadtree.evChanged.subscribe(self._sigQuadtreeChanged.emit)
        self.quadtree.evConfigChanged.subscribe(self.sigQuadtreeConfigChanged.emit)

        self.covariance.evChanged.subscribe(self.sigCovarianceChanged.emit)
        self.covariance.evConfigChanged.subscribe(self.sigCovarianceConfigChanged.emit)

    @QtCore.pyqtSlot(str)
    def exportWeightMatrix(self, filename):
        t0 = datetime.now()
        quadtree = self.quadtree
        covariance = self.covariance

        def progress_func():
            return covariance.finished_combinations

        ncombinations = quadtree.nleaves * (quadtree.nleaves + 1) / 2

        self.sigProgressStarted.emit(
            (
                'Calculating <span style="font-family: monospace">'
                "Covariance.weight_matrix</span>, this can take a few minutes...",
                ncombinations,
                progress_func,
            )
        )

        self.scene.covariance.export_weight_matrix(filename)
        self.sigProgressFinished.emit()
        self.sigCalculateWeightMatrixFinished.emit(datetime.now() - t0)

    @QtCore.pyqtSlot()
    def calculateWeightMatrix(self):
        t0 = datetime.now()
        quadtree = self.quadtree
        covariance = self.covariance

        def progress_func():
            return covariance.finished_combinations

        ncombinations = quadtree.nleaves * (quadtree.nleaves + 1) / 2

        self.sigProgressStarted.emit(
            (
                'Calculating <span style="font-family: monospace">'
                "Covariance.weight_matrix</span>,"
                " this can take a few minutes...",
                ncombinations,
                progress_func,
            )
        )

        self.scene.covariance.weight_matrix
        self.sigProgressFinished.emit()
        self.sigCalculateWeightMatrixFinished.emit(datetime.now() - t0)

    @QtCore.pyqtSlot(str)
    def importFile(self, filename):
        self.sigProgressStarted.emit(("Importing scene...",))
        self.setScene(Scene.import_data(filename))
        self.sigProgressFinished.emit()

    @QtCore.pyqtSlot(str)
    def loadFile(self, filename):
        self.sigProgressStarted.emit(("Loading scene...",))
        self.setScene(Scene.load(filename))
        self.sigProgressFinished.emit()

    @QtCore.pyqtSlot(str)
    def loadConfig(self, filename):
        self.scene.load_config(filename)

    def download_progress(self, context_str, status):
        progress = self.spool.progress

        progress.setWindowTitle("Downloading...")
        progress.setLabelText(context_str)
        progress.setMaximum(status.get("ntotal_bytes_all_files", 0))
        progress.setValue(status.get("nread_bytes_all_files", 0))
        QtCore.QCoreApplication.processEvents()

        if progress.isHidden():
            progress.show()
            QtCore.QCoreApplication.processEvents()

        if status["finished"]:
            progress.reset()


class QSceneQuadtreeProxy(QtCore.QObject):
    def __init__(self, scene_proxy):
        QtCore.QObject.__init__(self, scene_proxy)
        self.scene_proxy = scene_proxy

    @QtCore.pyqtSlot(float)
    def setEpsilon(self, value):
        self.scene_proxy.quadtree.epsilon = value

    @QtCore.pyqtSlot(float)
    def setNanFraction(self, value):
        self.scene_proxy.quadtree.nan_allowed = value

    @QtCore.pyqtSlot(float)
    def setTileMaximum(self, value):
        self.scene_proxy.quadtree.tile_size_max = value

    @QtCore.pyqtSlot(float)
    def setTileMinimum(self, value):
        self.scene_proxy.quadtree.tile_size_min = value



#!/usr/bin/env python3
import os
import platform
import tempfile
from distutils.sysconfig import get_python_inc
from os.path import join as pjoin

from setuptools import Extension, setup

try:
    import numpy
except ImportError:

    class numpy:
        def __init__(self):
            ...

        @classmethod
        def get_include(cls):
            return ""


def _have_openmp():
    """Check  whether the default compiler supports OpenMP.
    This routine is adapted from pynbody // yt.
    Thanks to Nathan Goldbaum and Andrew Pontzen.
    """
    import distutils.sysconfig
    import shutil
    import subprocess

    tmpdir = tempfile.mkdtemp(prefix="kite")
    compiler = os.environ.get("CC", distutils.sysconfig.get_config_var("CC")).split()[0]

    # Attempt to compile a test script.
    # See http://openmp.org/wp/openmp-compilers/
    tmpfile = pjoin(tmpdir, "check_openmp.c")
    with open(tmpfile, "w") as f:
        f.write(
            """
#include <omp.h>
#include <stdio.h>
int main() {
    #pragma omp parallel
    printf("Hello from thread %d", omp_get_thread_num());
}
"""
        )

    try:
        with open(os.devnull, "w") as fnull:
            exit_code = subprocess.call(
                [compiler, "-fopenmp", "-o%s" % pjoin(tmpdir, "check_openmp"), tmpfile],
                stdout=fnull,
                stderr=fnull,
            )
    except OSError:
        exit_code = 1
    finally:
        shutil.rmtree(tmpdir)

    if exit_code == 0:
        print("Continuing your build using OpenMP...\n")
        return True

    import multiprocessing

    if multiprocessing.cpu_count() > 1:
        print(
            """WARNING
OpenMP support is not available in your default C compiler, even though
your machine has more than one core available.
Some routines in kite are parallelized using OpenMP and these will
only run on one core with your current configuration.
"""
        )
        if platform.uname()[0] == "Darwin":
            print(
                """Since you are running on Mac OS, it's likely that the problem here
is Apple's Clang, which does not support OpenMP at all. The easiest
way to get around this is to download the latest version of gcc from
here: http://hpc.sourceforge.net. After downloading, just point the
CC environment variable to the real gcc and OpenMP support should
get enabled automatically. Something like this -
sudo tar -xzf /path/to/download.tar.gz /
export CC='/usr/local/bin/gcc'
python setup.py clean
python setup.py build
"""
            )
    print("Continuing your build without OpenMP...\n")
    return False


if _have_openmp():

    omp_arg = ["-fopenmp"]
    omp_lib = ["-lgomp"]

    if platform.uname()[0] == "Darwin":
        gomp_lib = os.environ.get("GOMPLIB", None)
        if gomp_lib:
            omp_lib.insert(0, "-L{}".format(gomp_lib))
            omp_lib.append("-Wl,-rpath,{}".format(gomp_lib))

        else:
            print(
                """Found the gcc compiler on MacOS but cannot find the
OpenMP libraries path at environment variable GOMPLIB.
"""
            )
            print("Continuing your build without OpenMP...\n")
            omp_arg = []
            omp_lib = []

else:
    omp_arg = []
    omp_lib = []


setup(
    ext_modules=[
        Extension(
            "kite.covariance_ext",
            sources=[pjoin("kite/ext", "covariance.c")],
            include_dirs=[numpy.get_include(), get_python_inc()],
            extra_compile_args=omp_arg,
            extra_link_args=omp_lib,
            language="c",
        ),
        Extension(
            "kite.sources.disloc_ext",
            sources=[pjoin("kite/sources/ext", "disloc.c")],
            include_dirs=[numpy.get_include(), get_python_inc()],
            extra_compile_args=omp_arg,
            extra_link_args=omp_lib,
            language="c",
        ),
    ]
)



import numpy as np
from pyrocko import gf
from pyrocko.guts import Float, Int, String

from .base import SandboxSource, SandboxSourceRectangular, SourceProcessor

d2r = np.pi / 180.0
r2d = 180.0 / np.pi
km = 1e3
km3 = 1e9


class PyrockoSource(object):
    def parametersUpdated(self):
        self.updatePyrockoSource()
        SandboxSource.parametersUpdated(self)

    def updatePyrockoSource(self):
        for arg, value in self._src_args.items():
            self.pyrocko_source.__setattr__(arg, value)


class PyrockoRectangularSource(SandboxSourceRectangular, PyrockoSource):
    """Classical Haskell source model modified for bilateral rupture.
    See :class:`pyrocko.gf.seismosizer.RectangularSource`.
    """

    __implements__ = "pyrocko"

    decimation_factor = Int.T(
        optional=True, default=10, help="Sub-source decimation factor."
    )
    store_dir = String.T(help="Pyrocko GF Store path")

    parametersUpdated = PyrockoSource.parametersUpdated

    def __init__(self, *args, **kwargs):
        SandboxSourceRectangular.__init__(self, *args, **kwargs)
        self.pyrocko_source = gf.RectangularSource(**self._src_args)

    @property
    def _src_args(self):
        return {
            "lat": 0.0,
            "lon": 0.0,
            "north_shift": self.northing,
            "east_shift": self.easting,
            "depth": self.depth,
            "length": self.length,
            "width": self.width,
            "strike": self.strike,
            "dip": self.dip,
            "rake": self.rake,
            "slip": self.slip,
            "decimation_factor": self.decimation_factor,
            "anchor": "top",
        }


class PyrockoMomentTensor(SandboxSource, PyrockoSource):
    """A moment tensor point source.

    See :class:`pyrocko.gf.seismosizer.MomentTensor`.
    """

    __implements__ = "pyrocko"

    store_dir = String.T(help="Pyrocko GF Store path")
    mnn = Float.T(default=1.0, help="north-north component of moment tensor in [Nm]")
    mee = Float.T(default=1.0, help="east-east component of moment tensor in [Nm]")
    mdd = Float.T(default=1.0, help="down-down component of moment tensor in [Nm]")
    mne = Float.T(default=0.0, help="north-east component of moment tensor in [Nm]")
    mnd = Float.T(default=0.0, help="north-down component of moment tensor in [Nm]")
    med = Float.T(default=0.0, help="east-down component of moment tensor in [Nm]")

    parametersUpdated = PyrockoSource.parametersUpdated

    def __init__(self, *args, **kwargs):
        SandboxSource.__init__(self, *args, **kwargs)
        self.pyrocko_source = gf.MTSource(**self._src_args)

    @property
    def _src_args(self):
        return {
            "lat": 0.0,
            "lon": 0.0,
            "north_shift": self.northing,
            "east_shift": self.easting,
            "depth": self.depth,
            "mnn": self.mnn,
            "mee": self.mee,
            "mdd": self.mdd,
            "mne": self.mne,
            "mnd": self.mnd,
            "med": self.med,
        }


class PyrockoDoubleCouple(SandboxSource, PyrockoSource):
    """A double-couple point source.

    See :class:`pyrocko.gf.seismosizer.DCSource`.
    """

    __implements__ = "pyrocko"

    strike = Float.T(
        default=0.0, help="strike direction in [deg], measured clockwise from north"
    )
    magnitude = Float.T(
        default=6.0, help="moment magnitude Mw as in [Hanks and Kanamori, 1979]"
    )
    dip = Float.T(
        default=90.0, help="dip angle in [deg], measured downward from horizontal"
    )
    rake = Float.T(
        default=0.0,
        help="rake angle in [deg], "
        "measured counter-clockwise from right-horizontal "
        "in on-plane view",
    )
    store_dir = String.T(help="Pyrocko GF Store path")

    parametersUpdated = PyrockoSource.parametersUpdated

    def __init__(self, *args, **kwargs):
        SandboxSource.__init__(self, *args, **kwargs)
        self.pyrocko_source = gf.DCSource(**self._src_args)

    @property
    def moment(self):
        return self.pyrocko_source.moment

    @property
    def _src_args(self):
        return {
            "lat": 0.0,
            "lon": 0.0,
            "north_shift": self.northing,
            "east_shift": self.easting,
            "depth": self.depth,
            "magnitude": self.magnitude,
            "strike": self.strike,
            "dip": self.dip,
            "rake": self.rake,
        }


class PyrockoRingfaultSource(SandboxSource, PyrockoSource):
    """A ring fault with vertical doublecouples.

    See :class:`pyrocko.gf.seismosizer.RingfaultSource`.
    """

    __implements__ = "pyrocko"

    store_dir = String.T(help="Pyrocko GF Store path")
    diameter = Float.T(default=1.0, help="diameter of the ring in [m]")
    sign = Float.T(default=1.0, help="inside of the ring moves up (+1) or down (-1)")
    strike = Float.T(
        default=0.0,
        help="strike direction of the ring plane, clockwise from north," " in [deg]",
    )
    dip = Float.T(
        default=0.0, help="dip angle of the ring plane from horizontal in [deg]"
    )
    npointsources = Int.T(default=8, help="number of point sources to use")
    magnitude = Float.T(
        default=6.0, help="moment magnitude Mw as in [Hanks and Kanamori, 1979]"
    )

    parametersUpdated = PyrockoSource.parametersUpdated

    def __init__(self, *args, **kwargs):
        SandboxSource.__init__(self, *args, **kwargs)
        PyrockoSource.__init__(self)
        self.pyrocko_source = gf.RingfaultSource(**self._src_args)

    @property
    def _src_args(self):
        return {
            "lat": 0.0,
            "lon": 0.0,
            "north_shift": self.northing,
            "east_shift": self.easting,
            "depth": self.depth,
            "diameter": self.diameter,
            "strike": self.strike,
            "dip": self.dip,
            "magnitude": self.magnitude,
            "npointsources": self.npointsources,
        }


class PyrockoVLVDSource(SandboxSource, PyrockoSource):
    """A ring fault with vertical doublecouples.

    See :class:`pyrocko.gf.seismosizer.VLVDSource`.
    """

    __implements__ = "pyrocko"

    store_dir = String.T(help="Pyrocko GF Store path")
    volume_change = Float.T(default=0.25, help="Volume change in [km^3]")
    azimuth = Float.T(
        default=0.0, help="azimuth direction of CLVD, clockwise from north," " in [deg]"
    )
    dip = Float.T(default=90.0, help="dip angle of the CLVD from horizontal in [deg]")
    clvd_moment = Float.T(default=3e18, help="Moment in [Nm] of the CLVD contribution")

    parametersUpdated = PyrockoSource.parametersUpdated

    def __init__(self, *args, **kwargs):
        SandboxSource.__init__(self, *args, **kwargs)
        PyrockoSource.__init__(self)
        self.pyrocko_source = gf.VLVDSource(**self._src_args)

    @property
    def _src_args(self):
        return {
            "lat": 0.0,
            "lon": 0.0,
            "north_shift": self.northing,
            "east_shift": self.easting,
            "depth": self.depth,
            "volume_change": self.volume_change * km3,
            "azimuth": self.azimuth,
            "dip": self.dip,
            "clvd_moment": self.clvd_moment,
        }


class PyrockoProcessor(SourceProcessor):

    __implements__ = "pyrocko"

    def __init__(self, *args):
        SourceProcessor.__init__(self, *args)
        self.engine = gf.LocalEngine()

    def process(self, sources, sandbox, nthreads=0):
        result = {
            "processor_profile": dict(),
            "displacement.n": np.zeros(sandbox.frame.npixel),
            "displacement.e": np.zeros(sandbox.frame.npixel),
            "displacement.d": np.zeros(sandbox.frame.npixel),
        }

        coords = sandbox.frame.coordinatesMeter

        target = gf.StaticTarget(
            lats=np.full(sandbox.frame.npixel, sandbox.frame.llLat),
            lons=np.full(sandbox.frame.npixel, sandbox.frame.llLon),
            east_shifts=coords[:, 0],
            north_shifts=coords[:, 1],
            interpolation="nearest_neighbor",
        )

        store_dirs = set([src.store_dir for src in sources])
        for store_dir in store_dirs:
            self.engine.store_dirs = [store_dir]

            talpa_sources = [src for src in sources if src.store_dir == store_dir]

            pyr_sources = [src.pyrocko_source for src in talpa_sources]

            for src in sources:
                src.regularize()

            try:
                res = self.engine.process(pyr_sources, [target], nthreads=nthreads)

            except Exception as e:
                self._log.error(
                    "Could not execute pyrocko.gf.LocalEngine.process! \n"
                    "LocalEngine Exception: %s" % e
                )
                continue

            for ires, static_res in enumerate(res.static_results()):
                result["displacement.n"] += static_res.result["displacement.n"]
                result["displacement.e"] += static_res.result["displacement.e"]
                result["displacement.d"] += static_res.result["displacement.d"]

                talpa_sources[ires]._cached_result = static_res.result

        for src in sources:
            if src._cached_result is None:
                continue
            self._log.debug("Using cached displacement for %s" % src.__class__.__name__)
            result["displacement.n"] += src._cached_result["displacement.n"]
            result["displacement.e"] += src._cached_result["displacement.e"]
            result["displacement.d"] += src._cached_result["displacement.d"]

        return result



from .compound_sources import *  # noqa
from .okada import *  # noqa
from .pyrocko_gf import *  # noqa

__sources__ = [
    OkadaSource,
    PyrockoMomentTensor,
    PyrockoRectangularSource,  # noqa
    EllipsoidSource,
    PointCompoundSource,
]  # noqa



#!/usr/bin/env python3
import pkg_resources

from .covariance import Covariance  # noqa
from .quadtree import Quadtree  # noqa
from .sandbox_scene import SandboxScene, TestSandboxScene  # noqa
from .scene import Scene, TestScene, read  # noqa

__version__ = pkg_resources.get_distribution("kite").version



import numpy as np
from pyrocko.guts import Bool

from .plugin import Plugin, PluginConfig


class DerampConfig(PluginConfig):

    demean = Bool.T(optional=True, default=True)


class Deramp(Plugin):
    def __init__(self, scene, config=None):
        self.scene = scene
        self.config = config or DerampConfig()

        self._log = scene._log.getChild("Deramp")

    def get_ramp_coefficients(self, displacement):
        """Fit plane through the displacement data.

        :returns: Mean of the displacement and slopes in easting coefficients
            of the fitted plane. The array hold
            ``[offset_e, offset_n, slope_e, slope_n]``.
        :rtype: :class:`numpy.ndarray`
        """
        scene = self.scene
        msk = np.isfinite(displacement)
        displacement = displacement[msk]

        coords = scene.frame.coordinates[msk.flatten()]

        # Add ones for the offset
        coords = np.hstack((np.ones_like(coords), coords))

        coeffs, res, _, _ = np.linalg.lstsq(coords, displacement, rcond=None)

        return coeffs

    def set_demean(self, demean):
        assert isinstance(demean, bool)
        self.config.demean = demean
        self.update()

    def apply(self, displacement):
        """Fit a plane onto the displacement data and subtract it

        :param demean: Demean the displacement
        :type demean: bool
        :param inplace: Replace data of the scene (default: True)
        :type inplace: bool

        :return: ``None`` if ``inplace=True`` else a new Scene
        :rtype: ``None`` or :class:`~kite.Scene`
        """
        self._log.debug("De-ramping scene")
        coeffs = self.get_ramp_coefficients(displacement)
        coords = self.scene.frame.coordinates

        ramp = coeffs[2:] * coords
        if self.config.demean:
            ramp += coeffs[:2]

        ramp = ramp.sum(axis=1).reshape(displacement.shape)

        displacement -= ramp
        return displacement



import logging
import time

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.image import AxesImage

from kite.util import Subject

__all__ = ["ScenePlot", "QuadtreePlot", "CovariancePlot"]

_DEFAULT_IMSHOW = {"cmap": "RdBu", "aspect": "equal"}

_VALID_COMPONENTS = {
    "displacement": "LOS Displacement",
    "theta": "LOS Theta",
    "phi": "LOS Phi",
    "cartesian.dE": "Displacement dE",
    "cartesian.dN": "Displacement dN",
    "cartesian.dU": "Displacement dU",
}


class Plot2D(object):
    """Base class for matplotlib 2D plots"""

    def __init__(self, scene, **kwargs):
        self.evPlotChanged = Subject()
        self._scene = scene
        self._data = None

        self.fig = None
        self.ax = None
        self._show_plt = False
        self._colormap_symmetric = True

        self.title = "unnamed"

        self._log = logging.getLogger(self.__class__.__name__)

    def __call__(self, *args, **kwargs):
        return self.plot(*args, **kwargs)

    def setCanvas(self, **kwargs):
        """Set canvas to plot in

        :param figure: Matplotlib figure to plot in
        :type figure: :py:class:`matplotlib.Figure`
        :param axes: Matplotlib axes to plot in
        :type axes: :py:class:`matplotlib.Axes`
        :raises: TypeError
        """
        axes = kwargs.get("axes", None)
        figure = kwargs.get("figure", None)

        if isinstance(axes, plt.Axes):
            self.fig, self.ax = axes.get_figure(), axes
            self._show_plt = False
        elif isinstance(figure, plt.Figure):
            self.fig, self.ax = figure, figure.gca()
            self._show_plt = False
        elif axes is None and figure is None and self.fig is None:
            self.fig, self.ax = plt.subplots(1, 1)
            self._show_plt = True
        else:
            raise TypeError(
                "axes has to be of type matplotlib.Axes. "
                "figure has to be of type matplotlib.Figure"
            )
        self.image = AxesImage(self.ax)
        self.ax.add_artist(self.image)

    @property
    def data(self):
        """Data passed to matplotlib.image.AxesImage"""
        return self._data

    @data.setter
    def data(self, value):
        self._data = value
        self.image.set_data(self.data)
        self.colormapAdjust()

    @data.getter
    def data(self):
        if self._data is None:
            return np.zeros((50, 50))
        return self._data

    def _initImagePlot(self, **kwargs):
        """Initiate the plot

        :param figure: Matplotlib figure to plot in
        :type figure: :py:class:`matplotlib.Figure`
        :param axes: Matplotlib axes to plot in
        :type axes: :py:class:`matplotlib.Axes`
        """
        self.setCanvas(**kwargs)

        self.setColormap(kwargs.get("cmap", "RdBu"))
        self.colormapAdjust()

        self.ax.set_xlim((0, self._scene.frame.E.size))
        self.ax.set_ylim((0, self._scene.frame.N.size))
        self.ax.set_aspect("equal")
        self.ax.invert_yaxis()

        self.ax.set_title(self.title)

        def close_figure(ev):
            self.fig = None
            self.ax = None

        try:
            self.fig.canvas.mpl_connect("close_event", close_figure)
        except Exception:
            pass

    def plot(self, **kwargs):
        """Placeholder in prototype class

        :param figure: Matplotlib figure to plot in
        :type figure: :py:class:`matplotlib.Figure`
        :param axes: Matplotlib axes to plot in
        :type axes: :py:class:`matplotlib.Axes`
        :param **kwargs: kwargs are passed into `plt.imshow`
        :type **kwargs: dict
        :raises: NotImplemented
        """
        raise NotImplementedError()
        self._initImagePlot(**kwargs)
        if self._show_plt:
            plt.show()

    def _updateImage(self):
        self.image.set_data(self.data)

    def setColormap(self, cmap="RdBu"):
        """Set matplotlib colormap

        :param cmap: matplotlib colormap name, defaults to 'RdBu'
        :type cmap: str, optional
        """
        self.image.set_cmap(cmap)
        self.evPlotChanged.notify()

    def colormapAdjust(self):
        """Set colormap limits automatically

        :param symmetric: symmetric colormap around 0, defaults to True
        :type symmetric: bool, optional
        """
        vmax = np.nanmax(self.data)
        vmin = np.nanmin(self.data)
        self.colormap_limits = (vmin, vmax)

    @property
    def colormap_symmetric(self):
        return self._colormap_symmetric

    @colormap_symmetric.setter
    def colormap_symmetric(self, value):
        self._colormap_symmetric = value
        self.colormapAdjust()

    @property
    def colormap_limits(self):
        return self.image.get_clim()

    @colormap_limits.setter
    def colormap_limits(self, limits):
        if not isinstance(limits, tuple):
            raise AttributeError("Limits have to be a tuple (vmin, vmax)")
        vmin, vmax = limits

        if self.colormap_symmetric:
            _max = max(abs(vmin), abs(vmax))
            vmin, vmax = -_max, _max
        self.image.set_clim(vmin, vmax)

        self.evPlotChanged.notify()

    @staticmethod
    def _colormapsAvailable():
        return [  # ('Perceptually Uniform Sequential',
            #  ['viridis', 'inferno', 'plasma', 'magma']),
            # ('Sequential', ['Blues', 'BuGn', 'BuPu',
            #                 'GnBu', 'Greens', 'Greys', 'Oranges', 'OrRd',
            #                 'PuBu', 'PuBuGn', 'PuRd', 'Purples', 'RdPu',
            #               'Reds', 'YlGn', 'YlGnBu', 'YlOrBr', 'YlOrRd']),
            # ('Sequential (2)', ['afmhot', 'autumn', 'bone', 'cool',
            #                     'copper', 'gist_heat', 'gray', 'hot',
            #                     'pink', 'spring', 'summer', 'winter']),
            (
                "Diverging",
                [
                    "BrBG",
                    "bwr",
                    "coolwarm",
                    "PiYG",
                    "PRGn",
                    "RdBu",
                    "RdGy",
                    "RdYlBu",
                    "RdYlGn",
                    "Spectral",
                    "seismic",
                    "PuOr",
                ],
            ),
            (
                "Qualitative",
                [
                    "Accent",
                    "Dark2",
                    "Paired",
                    "Pastel1",
                    "Pastel2",
                    "Set1",
                    "Set2",
                    "Set3",
                ],
            ),
            # ('Miscellaneous', ['gist_earth', 'terrain', 'ocean',
            #                  'brg', 'CMRmap', 'cubehelix', 'gist_stern',
            #                    'gnuplot', 'gnuplot2', 'gist_ncar',
            #                    'nipy_spectral', 'jet', 'rainbow',
            #                    'gist_rainbow', 'hsv', 'flag', 'prism'])
        ]


class ScenePlot(Plot2D):
    """Plotting 2D displacements though Matplotlib"""

    def __init__(self, scene, **kwargs):
        Plot2D.__init__(self, scene, **kwargs)

        self.components_available = {
            "displacement": {
                "name": "LOS Displacement",
                "eval": lambda sc: sc.displacement,
            },
            "theta": {"name": "LOS Theta", "eval": lambda sc: sc.theta},
            "phi": {"name": "LOS Phi", "eval": lambda sc: sc.phi},
            "dE": {"name": "Displacement dE", "eval": lambda sc: sc.cartesian.dE},
            "dN": {"name": "Displacement dN", "eval": lambda sc: sc.cartesian.dN},
            "dU": {"name": "Displacement dU", "eval": lambda sc: sc.cartesian.dU},
        }

        self._component = "displacement"

    def plot(self, component="displacement", **kwargs):
        """Plots any component from Scene
                The following components are recognizes

                - 'cartesian.dE'
                - 'cartesian.dN'
                - 'cartesian.dU'
                - 'displacement'
                - 'phi'
                - 'theta'

                :param **kwargs: Keyword args forwarded to `matplotlib.plt.imshow()`
                :type **kwargs: {dict}
                :param component: Component to plot
        ['cartesian.dE', 'cartesian.dN', 'cartesian.dU',
        'displacement', 'phi', 'theta']
                :type component: {string}, optional
                :param axes: Axes instance to plot in, defaults to None
                :type axes: :py:class:`matplotlib.Axes`, optional
                :param figure: Figure instance to plot in, defaults to None
                :type figure: :py:class:`matplotlib.Figure`, optional
                :param **kwargs: kwargs are passed into `plt.imshow`
                :type **kwargs: dict
                :returns: Imshow instance
                :rtype: :py:class:`matplotlib.image.AxesImage`
                :raises: AttributeError
        """
        self._initImagePlot(**kwargs)
        self.component = component
        self.title = self.components_available[component]

        if self._show_plt:
            plt.show()

    @property
    def component(self):
        return self._component

    @component.setter
    def component(self, component):
        try:
            if component not in self.components_available.keys():
                raise AttributeError("Invalid component %s" % component)
            self.data = self.components_available[component]["eval"](self._scene)
        except AttributeError:
            raise AttributeError("Could not access component %s" % component)


class QuadtreePlot(Plot2D):
    """Plotting 2D Quadtrees though Matplotlib"""

    def __init__(self, quadtree, **kwargs):
        self._quadtree = quadtree

        self.components_available = {
            "mean": {
                "name": "Mean",
                "eval": lambda qt: qt.leaf_matrix_means,
            },
            "median": {
                "name": "Median",
                "eval": lambda qt: qt.leaf_matrix_medians,
            },
        }
        self._component = "mean"

        Plot2D.__init__(self, quadtree.scene)

    @property
    def component(self):
        return self._component

    @component.setter
    def component(self, component):
        try:
            if component not in self.components_available.keys():
                raise AttributeError("Invalid component %s" % component)
            self.data = self.components_available[component]["eval"](self._quadtree)
        except AttributeError:
            raise AttributeError("Could not access component %s" % component)

    def plot(self, **kwargs):
        """Plot current quadtree

        :param axes: Axes instance to plot in, defaults to None
        :type axes: [:py:class:`matplotlib.Axes`], optional
        :param figure: Figure instance to plot in, defaults to None
        :type figure: [:py:class:`matplotlib.Figure`], optional
        :param **kwargs: kwargs are passed into `plt.imshow`
        :type **kwargs: dict
        """
        self._initImagePlot(**kwargs)
        self.data = self._quadtree.leaf_matrix_means
        self.title = "Quadtree Means"

        self._addInfoText()

        if self._show_plt:
            plt.show()

    def _addInfoText(self):
        """Add number of leaves in self.ax"""
        self.ax.text(
            0.975,
            0.975,
            "%d Leafs" % len(self._quadtree.leaves),
            transform=self.ax.transAxes,
            ha="right",
            va="top",
        )

    def _update(self):
        t0 = time.time()

        self.ax.texts = []
        self._addInfoText()
        self.data = self._quadtree.leaf_matrix_means
        self.colormapAdjust()
        self.ax.draw_artist(self.image)

        self._log.info(
            "Redrew %d leaves [%0.8f s]"
            % (len(self._quadtree.leaves), time.time() - t0)
        )

    def interactive(self):
        """Simple interactive quadtree plot with matplot

        This is a relictic function
        """
        from matplotlib.widgets import Slider

        self._initImagePlot()

        def change_epsilon(e):
            self._quadtree.epsilon = e

        def close_figure(*args):
            self._quadtree.evChanged.unsubscribe(self._update)

        self.ax.set_position([0.05, 0.15, 0.90, 0.8])
        ax_eps = self.fig.add_axes([0.05, 0.1, 0.90, 0.03])

        self.data = self._quadtree.leaf_matrix_means
        self.title = "Quadtree Means - Interactive"

        self._addInfoText()

        epsilon = Slider(
            ax_eps,
            "Epsilon",
            self._quadtree.epsilon - 1.0 * self._quadtree.epsilon,
            self._quadtree.epsilon + 1.0 * self._quadtree.epsilon,
            valinit=self._quadtree.epsilon,
            valfmt="%1.3f",
        )

        # Catch events
        epsilon.on_changed(change_epsilon)
        self._quadtree.evChanged.subscribe(self._update)
        self.fig.canvas.mpl_connect("close_event", close_figure)

        plt.show()


class CovariancePlot(object):
    def __init__(self, covariance, *args, **kwargs):
        self._covariance = covariance
        self.variance = covariance.variance
        self.quadtree = covariance.quadtree
        self.scene = covariance.quadtree.scene
        self.fig = plt.figure(figsize=(11.692, 8.267))

        self.ax_noi = self.fig.add_subplot(221)
        self.ax_cov = self.fig.add_subplot(222)
        self.ax_svar = self.fig.add_subplot(224)
        self.ax_pow = self.fig.add_subplot(223)
        # self.ax_qud = self.fig.add_subplot(325)

        self.plotCovariance(self.ax_cov)
        self.plotSemivariogram(self.ax_svar)
        self.plotPowerspec(self.ax_pow)
        self.plotNoise(self.ax_noi)
        # self.plotQuadtreeWeight(self.ax_qud)

        # self.plotPowerfit()

        self.fig.subplots_adjust(
            left=0.05, bottom=0.075, right=0.95, top=0.95, wspace=0.2, hspace=0.25
        )

    def __call__(self):
        self.fig.show()

    def show(self):
        from matplotlib.pyplot import figure

        fig = figure(FigureClass=self)
        fig.show()

    def plotCovariance(self, ax):
        cov, d = self._covariance.getCovariance()
        var = np.empty_like(d)
        var.fill(self.variance)
        ax.plot(d, cov)

        # d_interp = np.linspace(d.min(), d.max()+10000., num=50)
        # ax.plot(d_interp, self._covariance.covariance(d_interp),
        #        label='Interpolated', ls='--')
        model = self._covariance.getModelFunction(d, *self._covariance.covariance_model)
        ax.plot(d, var, label="variance", ls="--")
        ax.plot(d, model, label="interpolated", ls="--")

        ax.legend(loc="best")
        ax.grid(alpha=0.4)
        ax.set_title("Covariogram")
        ax.set_xlabel("distance [$m$]")
        ax.set_ylabel("covariance [$m^2$]")

    def plotNoise(self, ax):
        noise_data = self._covariance.noise_data
        noise_coord = self._covariance.noise_coord

        ax.imshow(
            np.flipud(noise_data),
            aspect="equal",
            extent=(0, noise_coord[2], 0, noise_coord[3]),
        )

        ax.set_title("Noise Data")
        ax.set_xlabel("X [$m$]")
        ax.set_ylabel("Y [$m$]")

    def plotSemivariogram(self, ax):
        svar, d = self._covariance.structure_spatial
        var = np.empty_like(d)
        var.fill(self.variance)
        ax.plot(d, svar)
        ax.plot(d, var, label="variance", ls="--")

        ax.legend(loc=4)
        ax.grid(alpha=0.4)
        ax.set_title("Semi-Variogram")
        ax.set_xlabel("distance [$m$]")
        ax.set_ylabel("semi-variance [$m^2$]")

    def plotPowerspec(self, ax):
        (
            power_spec,
            k,
            dk,
            f_spec,
            k_x,
            k_y,
        ) = self._covariance.powerspecNoise1D()  # noiseSpectrum()

        # power_spec_x = np.mean(f_spec, axis=1)
        # power_spec_y = np.mean(f_spec, axis=0)

        # ax.plot(k_x[k_x > 0], power_spec_x[k_x > 0], label='$k_x$')
        # ax.plot(k_y[k_y > 0], power_spec_y[k_y > 0], label='$k_y$')
        ax.plot(k, power_spec, label="$k_{total}$")

        # ax.legend(loc=1)
        ax.grid(alpha=0.4, which="both")
        ax.set_title("Power Spectrum")
        ax.set_xlabel("wavenumber [$cycles/m$]")
        ax.set_xscale("log")
        ax.set_ylabel("power [$m^2$]")
        ax.set_yscale("log")

    def plotQuadtreeWeight(self, ax):
        # extent = (self._quadtree.frame.llE, self._quadtree.frame.urE,
        #           self._quadtree.frame.llN, self._quadtree.frame.urN)
        # cb = ax.imshow(self._quadtree.leaf_matrix_weights, aspect='equal',
        #                extent=extent)
        ax.set_xlabel("UTM X [$m$]")
        ax.set_ylabel("UTM Y [$m$]")

    def plotPowerfit(self):
        import scipy as sp

        def behaviour(k, a, b):
            return (k**a) / b

        def selectRegime(k, d1, d2):
            return np.logical_and(((1.0) / k) > d1, ((1.0) / k) < d2)

        def curve_fit(k, p_spec):
            return sp.optimize.curve_fit(
                behaviour,
                k,
                p_spec,
                p0=None,
                sigma=None,
                absolute_sigma=False,
                check_finite=True,
                bounds=(-np.inf, np.inf),
                method=None,
                jac=None,
            )

        def covar_analyt(p, k):
            ps = behaviour(k[k > 0], *p)
            cov = sp.fftpack.dct(ps, type=2, n=None, axis=-1, norm="ortho")
            d = np.arange(1, k[k > 0].size + 1) * self.scene.frame.dE
            return cov, d

        power_spec, k, f_spec, k_x, k_y = self._covariance.noiseSpectrum()
        # Regimes accord. to Hanssen, 2001
        reg1 = selectRegime(k_x, 2000.0, np.inf)
        reg2 = selectRegime(k_x, 500.0, 2000.0)
        reg3 = selectRegime(k_x, 10.0, 500.0)

        for i, r in enumerate([reg1, reg2, reg3]):
            p, cov = curve_fit(k[r], power_spec[r])
            self.ax_pow.plot(
                k[r],
                behaviour(k[r], *p),
                ls="--",
                lw=1.5,
                alpha=0.8,
                label="Fit regime %d" % (i + 1),
            )

            cov, d = covar_analyt(p, k)
            self.ax_cov.plot(
                d, cov, ls="--", lw=1.5, alpha=0.8, label="Fit regime %d" % (i + 1)
            )

        self.ax_cov.legend(loc=1)
        self.ax_pow.legend(loc=1)


class SyntheticNoisePlot(object):
    def __init__(self, covariance, *args, **kwargs):
        self._covariance = covariance
        self.noise_data = self._covariance.noise_data
        clim_max = np.nanmax(self.noise_data)
        clim_min = np.nanmin(self.noise_data)
        self.clim = min(abs(clim_max), abs(clim_min))
        self.fig = plt.figure(figsize=(6.692, 3.149))

        self.ax_noi = self.fig.add_subplot(121)
        self.ax_snoi = self.fig.add_subplot(122)

        self.plotSynthNoise(self.ax_snoi)
        self.plotNoise(self.ax_noi)
        # self.plotQuadtreeWeight(self.ax_qud)

        # self.plotPowerfit()

        self.fig.subplots_adjust(
            left=0.125, bottom=0.1, right=0.9, top=0.9, wspace=0.5, hspace=0.25
        )

    def __call__(self):
        self.fig.show()

    def show(self):
        from matplotlib.pyplot import figure

        fig = figure(FigureClass=self)
        fig.show()

    def plotNoise(self, ax):
        noise_coord = self._covariance.noise_coord

        im = ax.imshow(
            np.flipud(self.noise_data),
            aspect="equal",
            extent=(0, noise_coord[2], 0, noise_coord[3]),
        )

        ax.set_title("Noise Data")
        ax.set_xlabel("X [$m$]")
        ax.set_ylabel("Y [$m$]")
        cbar = plt.colorbar(im, ax=ax)
        cbar.set_label("LOS displacement [$m$]")
        im.set_clim(-self.clim, self.clim)

    def plotSynthNoise(self, ax):
        noise_shape = np.shape(self.noise_data)
        noise_data = self._covariance.syntheticNoise(noise_shape)
        noise_coord = self._covariance.noise_coord

        im = ax.imshow(
            np.flipud(noise_data),
            aspect="equal",
            extent=(0, noise_coord[2], 0, noise_coord[3]),
        )

        ax.set_title("Synthetic Noise")
        ax.set_xlabel("X [$m$]")
        ax.set_ylabel("Y [$m$]")
        cbar = plt.colorbar(im, ax=ax)
        cbar.set_label("LOS displacement [$m$]")
        im.set_clim(-self.clim, self.clim)


if __name__ == "__main__":
    from kite.scene import SceneSynTest

    sc = SceneSynTest.createGauss()



import matplotlib.pyplot as plt
import pytest

from kite.gacos import GACOSCorrection


@pytest.skip(reason="data is local")
def test_gacos():
    corr = GACOSCorrection()
    corr.load("/home/marius/Development/testing/kite/GACOS/20180826.ztd")

    grd = corr.grids[0]

    d = grd.get_corrections(
        grd.llLat, grd.llLon, -grd.dLat, grd.dLon, grd.rows, grd.cols
    )

    d = grd.get_corrections(
        grd.llLat, grd.llLon, -grd.dLat * 2, grd.dLon * 2, grd.rows // 2, grd.cols // 2
    )

    d = grd.get_corrections(
        grd.llLat + 0.01,
        grd.llLon + 0.01,
        -grd.dLat / 1.5,
        grd.dLon / 1.5,
        int(grd.rows * 1.0),
        int(grd.cols * 1.0),
    )

    plt.imshow(grd.data)
    plt.show()
    plt.imshow(d)
    plt.show()


if __name__ == "__main__":
    test_gacos()



import time
from hashlib import sha1

import numpy as np
from pyrocko import guts
from pyrocko import orthodrome as od

from .util import Subject, derampMatrix, property_cached


class QuadNode(object):
    """A node (or *tile*) in held by :class:`~kite.Quadtree`. Each node in the
    tree hold a back reference to the quadtree and scene to access

    :param llr: Lower left corner row in :attr:`kite.Scene.displacement`
        matrix.
    :type llr: int
    :param llc: Lower left corner column in :attr:`kite.Scene.displacement`
        matrix.
    :type llc: int
    :param length: Length of node in from ``llr, llc`` in both dimensions
    :type length: int
    :param id: Unique id of node
    :type id: str
    :param children: Node's children
    :type children: List of :class:`~kite.quadtree.QuadNode`
    """

    CORNERS = (0, 0), (0, 1), (1, 0), (1, 1)
    MIN_PIXEL_LENGTH_NODE = None

    def __init__(self, quadtree, displacement, llr, llc, length):
        self.children = []
        self.llr = int(llr)
        self.llc = int(llc)
        self.length = int(length)
        self._slice_rows = slice(self.llr, self.llr + self.length)
        self._slice_cols = slice(self.llc, self.llc + self.length)
        self.id = "node_%d-%d_%d" % (self.llr, self.llc, self.length)

        self.quadtree = quadtree
        self._displacement = displacement
        self.scene = quadtree.scene
        self.frame = quadtree.frame

    @property_cached
    def nan_fraction(self):
        """Fraction of NaN values within the tile
        :type: float
        """
        return float(np.sum(self.displacement_mask)) / self.displacement.size

    @property_cached
    def npixel(self):
        return self.displacement.size

    @property_cached
    def mean(self):
        """Mean displacement
        :type: float
        """
        return float(np.nanmean(self.displacement))

    @property_cached
    def median(self):
        """Median displacement
        :type: float
        """
        return float(np.nanmedian(self.displacement))

    @property_cached
    def std(self):
        """Standard deviation of displacement
        :type: float
        """
        return float(np.nanstd(self.displacement))

    @property_cached
    def var(self):
        """Variance of displacement
        :type: float
        """
        return float(np.nanvar(self.displacement))

    @property_cached
    def mean_px_var(self):
        """Variance of displacement
        :type: float
        """
        if self.displacement_px_var is not None:
            return float(np.nanmean(self.displacement_px_var))
        return None

    @property_cached
    def corr_median(self):
        """Standard deviation of node's displacement corrected for median
        :type: float
        """
        return float(np.nanstd(self.displacement - self.median))

    @property_cached
    def corr_mean(self):
        """Standard deviation of node's displacement corrected for mean
        :type: float
        """
        return float(np.nanstd(self.displacement - self.mean))

    @property_cached
    def corr_bilinear(self):
        """Standard deviation of node's displacement corrected for bilinear
            trend (2D)
        :type: float
        """
        return float(np.nanstd(derampMatrix(self.displacement)))

    @property
    def weight(self):
        """
        :getter: Absolute weight derived from :class:`kite.Covariance`
         - works on tree leaves only.
        :type: float
        """
        return float(self.scene.covariance.getLeafWeight(self))

    @property_cached
    def focal_point(self):
        """Node focal point in local coordinates respecting NaN values
        :type: tuple, float - (easting, northing)
        """
        E = float(np.mean(self.gridE.compressed()) + self.frame.dE / 2)
        N = float(np.mean(self.gridN.compressed()) + self.frame.dN / 2)
        return E, N

    @property_cached
    def focal_point_meter(self):
        """Node focal point in local coordinates respecting NaN values
        :type: tuple, float - (easting, northing)
        """
        E = float(np.mean(self.gridEmeter.compressed() + self.frame.dEmeter / 2))
        N = float(np.mean(self.gridNmeter.compressed() + self.frame.dNmeter / 2))
        return E, N

    @property_cached
    def displacement(self):
        """Displacement array, slice from :attr:`kite.Scene.displacement`
        :type: :class:`numpy.ndarray`
        """
        return self._displacement[self._slice_rows, self._slice_cols]

    @property_cached
    def displacement_masked(self):
        """Masked displacement,
            see :attr:`~kite.quadtree.QuadNode.displacement`
        :type: :class:`numpy.ndarray`
        """
        return np.ma.masked_array(
            self.displacement, self.displacement_mask, fill_value=np.nan
        )

    @property_cached
    def displacement_mask(self):
        """Displacement nan mask of
            :attr:`~kite.quadtree.QuadNode.displacement`
        :type: :class:`numpy.ndarray`, dtype :class:`numpy.bool`

        .. todo ::

            Faster to slice Scene.displacement_mask?
        """
        return np.isnan(self.displacement)

    @property_cached
    def displacement_px_var(self):
        """Displacement array, slice from :attr:`kite.Scene.displacement`
        :type: :class:`numpy.ndarray`
        """
        if self.scene.displacement_px_var is not None:
            return self.scene.displacement_px_var[self._slice_rows, self._slice_cols]
        return None

    @property_cached
    def phi(self):
        """Median Phi angle, see :class:`~kite.Scene`.
        :type: float
        """
        phi = self.scene.phi[self._slice_rows, self._slice_cols]
        return np.nanmedian(phi[~self.displacement_mask])

    @property_cached
    def theta(self):
        """Median Theta angle, see :class:`~kite.Scene`.
        :type: float
        """
        theta = self.scene.theta[self._slice_rows, self._slice_cols]
        return np.nanmedian(theta[~self.displacement_mask])

    @property
    def unitE(self):
        unitE = self.scene.los_rotation_factors[self._slice_rows, self._slice_cols, 1]
        return np.nanmedian(unitE[~self.displacement_mask])

    @property
    def unitN(self):
        unitN = self.scene.los_rotation_factors[self._slice_rows, self._slice_cols, 2]
        return np.nanmedian(unitN[~self.displacement_mask])

    @property
    def unitU(self):
        unitU = self.scene.los_rotation_factors[self._slice_rows, self._slice_cols, 0]
        return np.nanmedian(unitU[~self.displacement_mask])

    @property_cached
    def gridE(self):
        """Grid holding local east coordinates,
            see :attr:`kite.scene.Frame.gridE`.
        :type: :class:`numpy.ndarray`
        """
        return self.scene.frame.gridE[self._slice_rows, self._slice_cols]

    @property_cached
    def gridEmeter(self):
        """Grid holding local east coordinates,
            see :attr:`kite.scene.Frame.gridEmeter`.
        :type: :class:`numpy.ndarray`
        """
        return self.scene.frame.gridEmeter[self._slice_rows, self._slice_cols]

    @property_cached
    def gridN(self):
        """Grid holding local north coordinates,
            see :attr:`kite.scene.Frame.gridN`.
        :type: :class:`numpy.ndarray`
        """
        return self.scene.frame.gridN[self._slice_rows, self._slice_cols]

    @property_cached
    def gridNmeter(self):
        """Grid holding local north coordinates,
            see :attr:`kite.scene.Frame.gridNmeter`.
        :type: :class:`numpy.ndarray`
        """
        return self.scene.frame.gridNmeter[self._slice_rows, self._slice_cols]

    @property
    def llE(self):
        """
        :getter: Lower left east coordinate in local coordinates
            (*meters* or *degree*).
        :type: float
        """
        return self.scene.frame.E[self.llc]

    @property
    def llN(self):
        """
        :getter: Lower left north coordinate in local coordinates
            (*meter* or *degree*).
        :type: float
        """
        return self.scene.frame.N[self.llr]

    @property
    def urN(self):
        return self.llN + self.sizeN

    @property
    def urE(self):
        return self.llE + self.sizeE

    @property_cached
    def sizeE(self):
        """
        :getter: Size in eastern direction in *meters* or *degree*.
        :type: float
        """
        sizeE = self.length * self.scene.frame.dE
        if (self.llE + sizeE) > self.scene.frame.E.max():
            sizeE = self.scene.frame.E.max() - self.llE
        return sizeE

    @property_cached
    def sizeN(self):
        """
        :getter: Size in northern direction in *meters* or *degree*.
        :type: float
        """
        sizeN = self.length * self.scene.frame.dN
        if (self.llN + sizeN) > self.scene.frame.N.max():
            sizeN = self.scene.frame.N.max() - self.llN
        return sizeN

    def iterChildren(self):
        """Iterator over the all children.

        :yields: Children of it's own.
        :type: :class:`~kite.quadtree.QuadNode`
        """
        yield self
        if self.children is not None:
            for c in self.children:
                yield from c.iterChildren()

    def iterLeaves(self):
        """Iterator over the leaves, evaluating parameters from
        :class:`~kite.Quadtree` instance.

        :yields: Leafs fulfilling the tree's parameters.
        :type: :class:`~kite.quadtree.QuadNode`
        """
        if (
            (
                self.quadtree._corr_func(self) < self.quadtree.epsilon
                and not self.length > self.quadtree._tile_size_lim_px[1]
            )
            or self.children is None
            or (self.children[0].length < self.quadtree._tile_size_lim_px[0])
        ):
            yield self
        else:
            for c in self.children:
                yield from c.iterLeaves()

    def _iterSplitNode(self):
        if self.length == 1:
            yield None
        for nr, nc in self.CORNERS:
            n = QuadNode(
                self.quadtree,
                self._displacement,
                self.llr + self.length / 2 * nr,
                self.llc + self.length / 2 * nc,
                self.length / 2,
            )
            if n.displacement.size == 0 or np.all(n.displacement_mask):
                continue
            yield n

    def createTree(self):
        """Create the tree from a set of basenodes, ignited by
        :class:`~kite.Quadtree` instance. Evaluates :class:`~kite.Quadtree`
        correction method and :attr:`~kite.Quadtree.epsilon_min`.
        """
        if (
            self.quadtree._corr_func(self) > self.quadtree.epsilon_min
            or self.length >= 64
        ) and not self.length < self.MIN_PIXEL_LENGTH_NODE:
            # self.length > .1 * max(self.quadtree._data.shape): !! Expensive
            self.children = tuple(c for c in self._iterSplitNode())
            if len(self.children) == 0:
                self.children = None
            else:
                for c in self.children:
                    c.createTree()
        else:
            self.children = None


class QuadtreeConfig(guts.Object):
    """Quadtree configuration object holding essential parameters used to
    reconstruct a particular tree
    """

    correction = guts.StringChoice.T(
        choices=("mean", "median", "bilinear", "std"),
        default="median",
        help="Node correction for splitting, available methods "
        " ``['mean', 'median', 'bilinear', 'std']``",
    )
    epsilon = guts.Float.T(
        optional=True, help="Variance threshold when a node is split"
    )
    nan_allowed = guts.Float.T(default=0.9, help="Allowed NaN fraction per tile")
    tile_size_min = guts.Float.T(
        optional=True, help="Minimum allowed tile size in *meters* or *degree*"
    )
    tile_size_max = guts.Float.T(
        optional=True, help="Maximum allowed tile size in *meters* or *degree*"
    )
    leaf_blacklist = guts.List.T(
        optional=True, default=[], help="Blacklist of excluded leaves"
    )


class Quadtree(object):
    """Quadtree for irregular subsampling InSAR displacement data held in
    :py:class:`kite.scene.Scene`

    InSAR displacement scenes can hold a vast amount of data points,
    which is often highly redundant and unsuitably large for the use in
    inverse modeling. By subsampling and therefore decimating the data points
    systematically through a parametrized quadtree we can reduce the dataset
    without significant loss of displacement information. Quadtree subsampling
    keeps a high spatial resolution where displacement gradients are high and
    efficiently reduces data point density in regions with small displacement
    variations. The product is a manageable dataset size with good
    representation of the original data.

    The standard deviation from :attr:`kite.quadtree.QuadNode.displacement`
    is evaluated against different corrections:

        * ``mean``: Mean is subtracted
        * ``median``: Median is subtracted
        * ``bilinear``: A 2D detrend is applied to the node
        * ``std``:  Pure standard deviation without correction

    set through :func:`~kite.Quadtree.setCorrection`. If the standard deviation
    exceeds :attr:`~kite.Quadtree.epsilon` the node is split.

    The leaves can also be exported in a *CSV* format by
    :func:`~kite.Quadtree.export_csv`, or *GeoJSON* by
    :func:`~kite.Quadtree.export_geojson`.

    Controlling attributes are:

        * :attr:`~kite.Quadtree.epsilon`, RMS threshold
        * :attr:`~kite.Quadtree.nan_fraction`, allowed :attr:`numpy.nan` in
          node
        * :attr:`~kite.Quadtree.tile_size_max`, maximum node size in
            *meters* or *degree*
        * :attr:`~kite.Quadtree.tile_size_min`, minimum node size in
            *meter* or *degree*

    :attr:`~kite.Quadtree.leaves` hold the current tree's
    :class:`~kite.quadtree.QuadNode` 's.
    """

    _displacement_corrections = {
        "mean": ("Standard deviation around mean", lambda n: n.corr_mean),
        "median": ("Standard deviation around median", lambda n: n.corr_median),
        "bilinear": (
            "Standard deviation around bilinear detrended node",
            lambda n: n.corr_bilinear,
        ),
        "std": ("Standard deviation (std)", lambda n: n.std),
    }

    _norm_methods = {
        "mean": lambda n: n.mean,
        "median": lambda n: n.median,
        "weight": lambda n: n.weight,
    }

    def __init__(self, scene, config=None):
        self.evChanged = Subject()
        self.evConfigChanged = Subject()
        self._leaves = None
        self.scene = scene
        self.displacement = self.scene.displacement
        self.frame = self.scene.frame
        self._scene_state = None

        # Cached matrices
        self._leaf_matrix_means = np.empty_like(self.displacement)
        self._leaf_matrix_medians = np.empty_like(self.displacement)
        self._leaf_matrix_weights = np.empty_like(self.displacement)

        self._log = scene._log.getChild("Quadtree")
        self.setConfig(config or QuadtreeConfig())

        self.scene.evConfigChanged.subscribe(self.setConfig)
        # self.scene.evChanged.subscribe(self.reinitializeTree)

    def setConfig(self, config=None):
        """Sets and updated the config of the instance

        :param config: New config instance, defaults to configuration provided
                       by parent :class:`~kite.Scene`
        :type config: :class:`~kite.covariance.QuadtreeConfig`, optional
        """
        if config is None:
            config = self.scene.config.quadtree

        if self.scene.config.old_import:
            frame = self.scene.config.frame

            from pyrocko import orthodrome as od

            self._log.warning("Old format - converting quadtree configuration")

            dLat, dLon = od.ne_to_latlon(
                frame.llLat, frame.llLon, config.tile_size_max, config.tile_size_min
            )

            config.tile_size_min = dLon - frame.llLon
            config.tile_size_max = dLat - frame.llLat

        self.config = config
        self.setCorrection(self.config.correction)

        self.evConfigChanged.notify()

    def setCorrection(self, correction="mean"):
        """Set correction method calculating the standard deviation of
        instances :class:`~kite.quadtree.QuadNode` s

        The standard deviation from :attr:`kite.quadtree.QuadNode.displacement`
        is evaluated against different corrections:

        * ``mean``: Mean is subtracted
        * ``median``: Median is subtracted
        * ``bilinear``: A 2D detrend is applied to the node
        * ``std``:  Pure standard deviation without correction

        :param correction: Choose from methods
            ``mean_std, median_std, bilinear_std, std``
        :type correction: str
        :raises: AttributeError
        """
        if correction not in self._displacement_corrections.keys():
            raise AttributeError(
                "Method %s not in %s", correction, self._displacement_corrections
            )
        self._log.debug("Changing to split method '%s'", correction)

        self.config.correction = correction
        self._corr_func = self._displacement_corrections[correction][1]
        self.reinitializeTree()

    def ensureTree(self):
        if self._scene_state != self.scene.get_plugin_state_hash():
            self.reinitializeTree()

    def reinitializeTree(self):
        # Clearing cached properties through None
        self.leaf_center_distance = None
        self.nodes = None
        self.epsilon_min = None
        self._epsilon_init = None
        self.clearLeaves()
        self.epsilon = self.config.epsilon or self._epsilon_init

        self._initTree()
        if self.nleaves == 0:
            self._log.warning(
                "No leaves in default quadtree," " setting allowed_nan=1."
            )
            self.nan_allowed = 1.0

        self.evChanged.notify()

    def clearLeaves(self):
        """Clear cached leafs and properties"""
        self.leaves = None
        self.leaf_center_distance = None
        self.leaf_los_rotation_factors = None
        self.leaf_means = None
        self.leaf_medians = None

    @property
    def min_node_length_px(self):
        npx = max(self.frame.cols, self.frame.rows)
        return int(2 ** round(np.log(npx / 64)))

    def _initTree(self):
        QuadNode.MIN_PIXEL_LENGTH_NODE = (
            QuadNode.MIN_PIXEL_LENGTH_NODE or self.min_node_length_px
        )

        t0 = time.time()
        for b in self._base_nodes:
            b.createTree()

        self._scene_state = self.scene.get_plugin_state_hash()
        self._log.debug(
            "Tree created, %d nodes [%0.4f s]", self.nnodes, time.time() - t0
        )

    @property
    def epsilon(self):
        """Threshold for quadtree splitting its ``QuadNode``.

        The threshold is the maximum standard deviation of leaf mean,
        median or simply its values (see ''SetSplitMethod'') allowed to
        not further split a "QuadNode".

        :setter: Sets the epsilon/RMS threshold
        :getter: Returns the current epsilon
        :type: float
        """
        return self.config.epsilon

    @epsilon.setter
    def epsilon(self, value):
        value = float(value)
        if self.config.epsilon == value:
            return
        if value < self.epsilon_min:
            self._log.warning(
                "Epsilon is out of bounds [%0.6f], epsilon_min %0.6f",
                value,
                self.epsilon_min,
            )
            return
        self.clearLeaves()
        self.clearLeafBlacklist()
        self.config.epsilon = value

        self.evChanged.notify()

    @property_cached
    def _epsilon_init(self):
        """Initial epsilon for virgin tree creation"""
        return np.nanstd(self.displacement)

    @property_cached
    def epsilon_min(self):
        """Lowest allowed epsilon
        :type: float
        """
        return self._epsilon_init * 0.1

    @property
    def nan_allowed(self):
        """Fraction of allowed ``NaN`` values in quadtree leaves. If
        value is exceeded the leaf is kicked out entirely.

        :setter: Fraction  ``0. <= fraction <= 1``.
        :type: float
        """
        return self.config.nan_allowed

    @nan_allowed.setter
    def nan_allowed(self, value):
        if value > 1.0 or value <= 0.0:
            self._log.warning("NaN fraction must be 0. < nan_allowed <= 1.")
            return

        self.clearLeaves()
        self.clearLeafBlacklist()
        self.config.nan_allowed = value
        self.evChanged.notify()

    @property
    def tile_size_min(self):
        """Minimum allowed tile size in *meter*.
        Measured along long edge ``(max(dE, dN))``.
        Minimum tile size defaults to 1/20th of the largest dimension

        :getter: Returns the minimum allowed tile size
        :setter: Sets the minimum threshold
        :type: float
        """
        if self.config.tile_size_min is None:
            frame = self.scene.frame
            max_px = max(frame.shape)
            self.config.tile_size_min = max(frame.dE, frame.dN) * (max_px / 20)

        return self.config.tile_size_min

    @tile_size_min.setter
    def tile_size_min(self, value):
        if value > self.tile_size_max:
            self._log.warning("tile_size_min > tile_size_max is required!")
            return
        self.config.tile_size_min = value
        self._tileSizeChanged()

    @property
    def tile_size_max(self):
        """Maximum allowed tile size in *meter*.
        Measured along long edge ``(max(dE, dN))``
        Maximum tile size defaults to 1/5th of the largest dimension

        :getter: Returns the maximum allowed tile size
        :setter: Sets the maximum threshold
        :type: float
        """
        if self.config.tile_size_max is None:
            frame = self.scene.frame
            max_px = max(frame.shape)
            self.config.tile_size_max = max(frame.dE, frame.dN) * (max_px / 5)

        return self.config.tile_size_max

    @tile_size_max.setter
    def tile_size_max(self, value):
        if value < self.tile_size_min:
            self._log.warning("tile_size_min > tile_size_max is required")
            return
        self.config.tile_size_max = value
        self._tileSizeChanged()

    def _tileSizeChanged(self):
        self._tile_size_lim_px = None
        self.clearLeaves()
        self.clearLeafBlacklist()
        self.evChanged.notify()

    @property_cached
    def _tile_size_lim_px(self):
        dpx = max(self.scene.frame.dE, self.scene.frame.dN)
        return (round(self.tile_size_min / dpx), round(self.tile_size_max / dpx))

    @property_cached
    def nodes(self):
        """All nodes of the tree

        :getter: Get the list of nodes
        :type: list
        """
        return [n for b in self._base_nodes for n in b.iterChildren()]

    @property
    def nnodes(self):
        """
        :getter: Number of nodes of the built tree.
        :type: int
        """
        return len(self.nodes)

    def clearLeafBlacklist(self):
        self.config.leaf_blacklist = []

    def blacklistLeaves(self, leaves):
        """Blacklist a leaf and exclude it from the tree

        :param leaves: Leaf instances
        :type leaves: list
        """
        self.config.leaf_blacklist.extend(leaves)
        self._log.debug(
            "Blacklisted leaves: %s" % ", ".join(self.config.leaf_blacklist)
        )
        self.clearLeaves()
        self.evChanged.notify()

    @property_cached
    def leaves(self):
        """:getter: List of leaves for current configuration.
        :type: (list or :class:`~kite.quadtree.QuadNode` s)
        """
        t0 = time.time()
        leaves = []
        for b in self._base_nodes:
            leaves.extend(
                [
                    lf
                    for lf in b.iterLeaves()
                    if lf.nan_fraction < self.nan_allowed
                    and lf.id not in self.config.leaf_blacklist
                ]
            )
        self._log.debug(
            "Gathering leaves for epsilon %.4f (nleaves=%d) [%0.4f s]"
            % (self.epsilon, len(leaves), time.time() - t0)
        )
        return leaves

    @property
    def nleaves(self):
        """
        :getter: Number of leaves for current parametrisation.
        :type: int
        """
        return len(self.leaves)

    @property
    def leaf_mean_px_var(self):
        """
        :getter: Mean pixel variance in each quadtree,
            if :attr:`kite.Scene.displacement_px_var` is set.
        :type: :class:`numpy.ndarray`, size ``N``.
        """
        if self.scene.displacement_px_var is not None:
            return np.array([lf.mean_px_var for lf in self.leaves])
        return None

    @property_cached
    def leaf_means(self):
        """
        :getter: Leaf mean displacements from
            :attr:`kite.quadtree.QuadNode.mean`.
        :type: :class:`numpy.ndarray`, size ``N``.
        """
        return np.array([lf.mean for lf in self.leaves])

    @property_cached
    def leaf_medians(self):
        """
        :getter: Leaf median displacements from
            :attr:`kite.quadtree.QuadNode.median`.
        :type: :class:`numpy.ndarray`, size ``N``.
        """
        return np.array([lf.median for lf in self.leaves])

    @property
    def _leaf_focal_points(self):
        return np.array([lf._focal_point for lf in self.leaves])

    @property
    def leaf_focal_points(self):
        """
        :getter: Leaf focal points in local coordinates.
        :type: :class:`numpy.ndarray`, size ``(N, 2)``
        """
        return np.array([lf.focal_point for lf in self.leaves])

    @property
    def leaf_focal_points_meter(self):
        """
        :getter: Leaf focal points in meter.
        :type: :class:`numpy.ndarray`, size ``(N, 2)``
        """
        return np.array([lf.focal_point_meter for lf in self.leaves])

    @property
    def leaf_coordinates(self):
        """Synonym for :func:`Quadtree.leaf_focal_points`
        in easting/northing"""
        return self.leaf_focal_points

    @property_cached
    def leaf_center_distance(self):
        """
        :getter: Leaf distance to center point of the quadtree
        :type: :class:`numpy.ndarray`, size ``(N, 3)``
        """
        distances = np.empty((self.nleaves, 3))
        center = self.center_point
        distances[:, 0] = self.leaf_focal_points[:, 0] - center[0]
        distances[:, 1] = self.leaf_focal_points[:, 1] - center[1]
        distances[:, 2] = np.sqrt(distances[:, 1] ** 2 + distances[:, 1] ** 2)
        return distances

    @property
    def leaf_eastings(self):
        return self.leaf_coordinates[:, 0]

    @property
    def leaf_northings(self):
        return self.leaf_coordinates[:, 1]

    @property
    def leaf_phis(self):
        """
        :getter: Median leaf LOS phi angle. :attr:`kite.Scene.phi`
        :type: :class:`numpy.ndarray`, size ``(N)``
        """
        return np.array([lf.phi for lf in self.leaves])

    @property
    def leaf_thetas(self):
        """
        :getter: Median leaf LOS theta angle. :attr:`kite.Scene.theta`
        :type: :class:`numpy.ndarray`, size ``(N)``
        """
        return np.array([lf.theta for lf in self.leaves])

    @property_cached
    def leaf_los_rotation_factors(self):
        """
        :getter: Trigonometric factors for rotating displacement
            matrices towards LOS.
            See :attr:`kite.BaseScene.los_rotation_factors`
        :type: :class:`numpy.ndarray`, Nx3
        """
        los_factors = np.empty((self.nleaves, 3))
        los_factors[:, 0] = np.sin(self.leaf_thetas)
        los_factors[:, 1] = np.cos(self.leaf_thetas) * np.cos(self.leaf_phis)
        los_factors[:, 2] = np.cos(self.leaf_thetas) * np.sin(self.leaf_phis)
        return los_factors

    @property
    def leaf_matrix_means(self):
        """
        :getter: Leaf mean displacements casted to
            :attr:`kite.Scene.displacement`.
        :type: :class:`numpy.ndarray`, size ``(N, M)``
        """
        return self._getLeafsNormMatrix(self._leaf_matrix_means, method="mean")

    @property
    def leaf_matrix_medians(self):
        """
        :getter: Leaf median displacements casted to
            :attr:`kite.Scene.displacement`.
        :type: :class:`numpy.ndarray`, size ``(N, M)``
        """
        return self._getLeafsNormMatrix(self._leaf_matrix_medians, method="median")

    @property
    def leaf_matrix_weights(self):
        """
        :getter: Leaf weights casted to :attr:`kite.Scene.displacement`.
        :type: :class:`numpy.ndarray`, size ``(N, M)``
        """
        return self._getLeafsNormMatrix(self._leaf_matrix_weights, method="weight")

    def _getLeafsNormMatrix(self, array, method="median"):
        if method not in self._norm_methods.keys():
            raise AttributeError(
                "Method %s is not in %s" % (method, list(self._norm_methods.keys()))
            )

        array.fill(np.nan)
        for lf in self.leaves:
            array[lf._slice_rows, lf._slice_cols] = self._norm_methods[method](lf)
        array[self.scene.displacement_mask] = np.nan
        return array

    @property
    def center_point(self):
        return np.median(self.leaf_focal_points, axis=0)

    @property
    def reduction_efficiency(self):
        """This is measure for the reduction of the scene's full resolution
        over the quadtree.

        :getter: Quadtree efficiency as :math:`N_{full} / N_{leaves}`
        :type: float
        """
        return (self.scene.rows * self.scene.cols) / (
            self.nleaves if self.nleaves else 1
        )

    @property
    def reduction_rms(self):
        """The RMS error is defined between
        :attr:`~kite.Quadtree.leaf_matrix_means` and
        :attr:`kite.Scene.displacement`.

        :getter: The reduction RMS error
        :type: float
        """
        if np.all(np.isnan(self.leaf_matrix_means)):
            return np.inf
        return np.sqrt(
            np.nanmean((self.scene.displacement - self.leaf_matrix_means) ** 2)
        )

    @property_cached
    def _base_nodes(self):
        self._base_nodes = []
        init_length = np.power(
            2, np.ceil(np.log(np.min(self.displacement.shape)) / np.log(2))
        )
        nx, ny = np.ceil(np.array(self.displacement.shape) / init_length)
        self._log.debug("Creating %d base nodes", nx * ny)

        displacement = self.scene.displacement
        for ir in range(int(nx)):
            for ic in range(int(ny)):
                llr = ir * init_length
                llc = ic * init_length
                node = QuadNode(self, displacement, llr, llc, init_length)
                self._base_nodes.append(node)

        if len(self._base_nodes) == 0:
            raise AssertionError("Could not init base nodes.")
        return self._base_nodes

    @property_cached
    def plot(self):
        """Simple `matplotlib` illustration of the quadtree

        :type: :attr:`Quadtree.leaf_matrix_means`.
        """
        from kite.plot2d import QuadtreePlot

        return QuadtreePlot(self)

    def getStaticTarget(self):
        """Not Implemented"""
        raise NotImplementedError

    def getMPLRectangles(self):
        """
        Get the quadtree as a list of matplotlib rectangles.

        :returns: Rectangles for plotting
        :rtype: list of :class:`matplotlib.patcjes.Rectangle`
        """
        from matplotlib.patches import Rectangle

        rectangles = []
        for lf in self.leaves:
            r = Rectangle((lf.llE, lf.llN), lf.sizeE, lf.sizeN)
            rectangles.append(r)
        return rectangles

    def export_csv(self, filename):
        """ Exports the current quadtree leaves to ``filename`` in a
        *CSV* format

        The formatting is::

            # node_id, focal_point_E, focal_point_N, theta, phi, \
            mean_displacement, median_displacement, absolute_weight

        :param filename: export_csv to path
        :type filename: string
        """
        self._log.debug("Exporting Quadtree as to %s", filename)
        with open(filename, mode="w") as f:
            f.write(
                "# node_id, focal_point_E, focal_point_N, theta, phi,"
                " unitE, unitN, unitU,"
                " mean_displacement, median_displacement, absolute_weight\n"
            )
            for lf in self.leaves:
                f.write(
                    "{lf.id}, {lf.focal_point[0]}, {lf.focal_point[1]}, "
                    "{lf.theta}, {lf.phi}, {lf.unitE}, {lf.unitN}, {lf.unitU},"
                    " {lf.mean}, {lf.median}, {lf.weight}\n".format(lf=lf)
                )

    def export_geojson(self, filename):
        import geojson

        self._log.debug("Exporting GeoJSON Quadtree to %s", filename)
        features = []

        for lf in self.leaves:
            llN, llE, urN, urE = (lf.llN, lf.llE, lf.urN, lf.urE)

            if self.frame.isDegree():
                llN += self.frame.llLat
                llE += self.frame.llLon
                urN += self.frame.llLat
                urE += self.frame.llLon

            coords = np.array(
                [(llN, llE), (llN, urE), (urN, urE), (urN, llE), (llN, llE)]
            )

            if self.frame.isMeter():
                coords = od.ne_to_latlon(self.frame.llLat, self.frame.llLon, *coords.T)
                coords = np.array(coords).T

            coords = coords[:, [1, 0]].tolist()

            feature = geojson.Feature(
                geometry=geojson.Polygon(coordinates=[coords]),
                id=lf.id,
                properties={
                    "mean": float(lf.mean),
                    "median": float(lf.median),
                    "std": float(lf.std),
                    "var": float(lf.var),
                    "phi": float(lf.phi),
                    "theta": float(lf.theta),
                    "unitE": float(lf.unitE),
                    "unitN": float(lf.unitN),
                    "unitU": float(lf.unitU),
                },
            )
            features.append(feature)

        collection = geojson.FeatureCollection(features)
        with open(filename, "w") as f:
            geojson.dump(collection, f)

    def get_state_hash(self):
        sha = sha1()
        sha.update(str(self.config).encode())
        return sha.digest().hex()


__all__ = ["Quadtree", "QuadtreeConfig"]


if __name__ == "__main__":
    from kite.scene import SceneSynTest

    sc = SceneSynTest.createGauss(2000, 2000)

    for e in np.linspace(0.1, 0.00005, num=30):
        sc.quadtree.epsilon = e
    # qp = Plot2DQuadTree(qt, cmap='spectral')
    # qp.plot()



import numpy as np

from kite import Scene


def test_deramp():
    c = np.arange(20, dtype=float)
    E, N = np.meshgrid(c, c)

    displ = (-3 + 5.4 * E) + (10 + 2.5 * N)
    sc = Scene(displacement=displ, llLat=0, llLon=0.0, dLat=0.3, dLon=0.3)

    coeffs = sc.deramp.get_ramp_coefficients(sc.displacement)
